@@ -1,4 +1,4 @@
-from h5py import VirtualLayout, VirtualSource, Dataset
+from h5py import VirtualLayout, VirtualSource, Dataset, Group, __version__ as h5py_version
 from h5py._hl.vds import VDSmap
 from h5py._hl.selections import select
 from h5py.h5i import get_name
@@ -139,10 +139,13 @@ def _recreate_raw_data(f, name, versions_to_delete, tmp=False):
 
         dataset = f['_version_data/versions'][version_name][name]
 
-        virtual_sources = dataset.virtual_sources()
-        slice_map = {spaceid_to_slice(i.vspace):
-                     spaceid_to_slice(i.src_space) for i in
-                     virtual_sources}
+        if dataset.is_virtual:
+            virtual_sources = dataset.virtual_sources()
+            slice_map = {spaceid_to_slice(i.vspace):
+                         spaceid_to_slice(i.src_space) for i in
+                         virtual_sources}
+        else:
+            slice_map = {}
         chunks_map[version_name].update(slice_map)
 
     chunks_to_keep = set().union(*[map.values() for map in
@@ -157,6 +160,21 @@ def _recreate_raw_data(f, name, versions_to_delete, tmp=False):
     chunks = ChunkSize(raw_data.chunks)
     new_shape = (len(chunks_to_keep)*chunks[0], *chunks[1:])
 
+    # See InMemoryDataset.fillvalue. In h5py3 variable length strings use None
+    # for the h5py fillvalue, but require a string fillvalue for NumPy.
+    def _get_np_fillvalue(data):
+        if data.fillvalue is not None:
+            return data.fillvalue
+        if data.dtype.metadata:
+             if 'vlen' in raw_data.dtype.metadata:
+                 if (h5py_version.startswith('3') and
+                     data.dtype.metadata['vlen'] == str):
+                     return bytes()
+                 return data.dtype.metadata['vlen']()
+             elif 'h5py_encoding' in raw_data.dtype.metadata:
+                 return data.dtype.type()
+        return np.zeros((), dtype=data.dtype)[()]
+
     new_raw_data = f['_version_data'][name].create_dataset(
         '_tmp_raw_data', shape=new_shape, maxshape=(None,)+chunks[1:],
         chunks=raw_data.chunks, dtype=raw_data.dtype,
@@ -167,7 +185,7 @@ def _recreate_raw_data(f, name, versions_to_delete, tmp=False):
         new_raw_data.attrs[key] = val
 
     r = raw_data[:]
-    n = np.full(new_raw_data.shape, new_raw_data.fillvalue, dtype=new_raw_data.dtype)
+    n = np.full(new_raw_data.shape, _get_np_fillvalue(raw_data), dtype=new_raw_data.dtype)
     raw_data_chunks_map = {}
     for new_chunk, chunk in zip(chunks.indices(new_shape), chunks_to_keep):
         # Shrink new_chunk to the size of chunk, in case chunk isn't a full
@@ -219,8 +237,8 @@ def _recreate_hashtable(f, name, raw_data_chunks_map, tmp=False):
 
 def _recreate_virtual_dataset(f, name, versions, raw_data_chunks_map, tmp=False):
     """
-    Recreate every virtual dataset `name` in the version versions according to
-    the new raw_data chunks.
+    Recreate every virtual dataset `name` in the versions `versions` according
+    to the new raw_data chunks in `raw_data_chunks_map`.
 
     Returns a dict mapping the chunks from the old raw dataset to the chunks
     in the new raw dataset. Chunks not in the mapping were deleted. If the
@@ -247,38 +265,42 @@ def _recreate_virtual_dataset(f, name, versions, raw_data_chunks_map, tmp=False)
         if not layout_has_sources:
             vs = VirtualSource('.', name=raw_data.name, shape=raw_data.shape, dtype=raw_data.dtype)
 
-        virtual_sources = dataset.virtual_sources()
-        for vmap in virtual_sources:
-            vspace, fname, dset_name, src_space = vmap
-            fname = fname.encode('utf-8')
-            assert fname == b'.', fname
+        # If a dataset has no data except for the fillvalue, it will not be virtual
+        if dataset.is_virtual:
+            virtual_sources = dataset.virtual_sources()
+            for vmap in virtual_sources:
+                vspace, fname, dset_name, src_space = vmap
+                fname = fname.encode('utf-8')
+                assert fname == b'.', fname
 
-            vslice = spaceid_to_slice(vspace)
-            src_slice = spaceid_to_slice(src_space)
-            if src_slice not in raw_data_chunks_map:
-                raise ValueError(f"Could not find the chunk for {vslice} ({src_slice} in the old raw dataset) for {name!r} in {version_name!r}")
-            new_src_slice = raw_data_chunks_map[src_slice]
+                vslice = spaceid_to_slice(vspace)
+                src_slice = spaceid_to_slice(src_space)
+                if src_slice not in raw_data_chunks_map:
+                    raise ValueError(f"Could not find the chunk for {vslice} ({src_slice} in the old raw dataset) for {name!r} in {version_name!r}")
+                new_src_slice = raw_data_chunks_map[src_slice]
 
-            # h5py 3.3 changed the VirtualLayout code. See
-            # https://github.com/h5py/h5py/pull/1905 and the code in
-            # create_virtual_dataset().
-            if layout_has_sources:
-                vs_sel = select(raw_data.shape, new_src_slice.raw, None)
-                layout_sel = select(dataset.shape, vslice.raw, None)
-                new_vmap = VDSmap(layout_sel.id, fname, dset_name, vs_sel.id)
-                layout.sources.append(new_vmap)
-            else:
-                layout[vslice.raw] = vs[new_src_slice.raw]
+                # h5py 3.3 changed the VirtualLayout code. See
+                # https://github.com/h5py/h5py/pull/1905 and the code in
+                # create_virtual_dataset().
+                if layout_has_sources:
+                    vs_sel = select(raw_data.shape, new_src_slice.raw, None)
+                    layout_sel = select(dataset.shape, vslice.raw, None)
+                    new_vmap = VDSmap(layout_sel.id, fname, dset_name, vs_sel.id)
+                    layout.sources.append(new_vmap)
+                else:
+                    layout[vslice.raw] = vs[new_src_slice.raw]
 
-        tmp_name = '_tmp_' + name
-        tmp_dataset = group.create_virtual_dataset(tmp_name, layout, fillvalue=dataset.fillvalue)
+        head, tail = pp.split(name)
+        tmp_name = '_tmp_' + tail
+        tmp_path = pp.join(head, tmp_name)
+        tmp_dataset = group.create_virtual_dataset(tmp_path, layout, fillvalue=dataset.fillvalue)
 
         for key, val in dataset.attrs.items():
             tmp_dataset.attrs[key] = val
 
         if not tmp:
             del group[name]
-            group.move(tmp_name, name)
+            group.move(tmp_path, name)
 
 def delete_versions(f, versions_to_delete):
     """
@@ -306,6 +328,7 @@ def delete_versions(f, versions_to_delete):
     def delete_dataset(name):
         if name == 'versions':
             return
+
         # Recreate the raw data.
         raw_data_chunks_map = _recreate_raw_data(f, name, versions_to_delete)
 
@@ -319,9 +342,20 @@ def delete_versions(f, versions_to_delete):
         # Recreate every virtual dataset in every kept version.
         _recreate_virtual_dataset(f, name, versions_to_keep, raw_data_chunks_map)
 
-    for name in version_data:
-        if name == 'versions':
-            continue
+    # We use this instead of version_data.visit(delete_dataset) because
+    # visit() has trouble with the groups being deleted from under it.
+    def _walk(g, prefix=''):
+        for name in g:
+            obj = g[name]
+            if isinstance(obj, Group):
+                if 'raw_data' in obj:
+                    to_delete.append(prefix + name)
+                else:
+                    _walk(obj, prefix + name + '/')
+
+    to_delete = []
+    _walk(version_data)
+    for name in to_delete:
         delete_dataset(name)
 
     for version in versions_to_delete:
